==> peach-cli-0.0.4.tar.gz: peach-cli-0.0.4/src/peach/jobs.py <==
from functools import cached_property
import logging
import re
from typing import Callable, Dict, List

import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions

from google.cloud import bigquery
import pandas as pd
import pyarrow as pa

from peach.constants import UpdateMethod
from peach.datastores import Datastore
from peach.table import Table


class Job:
    def update_table(
        self,
        datastore: Datastore,
        output: Table,
        inputs: Dict[str, Table],
    ) -> None:
        raise NotImplementedError()


class SqlJob(Job):

    def __init__(self, sql_file: str, update_method: UpdateMethod):
        self.sql_file = sql_file
        self.update_method = update_method

    @cached_property
    def sql_code(self) -> str:
        with open(self.sql_file, "r") as fo:
            return fo.read()

    def _substitute_table_names(
        self, id_to_locator_string_mapping: Dict[str, str], m: re.Match
    ) -> str:
        t = m.group(2)[5:-2]
        assert "." in t or "/" in t, f"Table does not specify schema: {t}"

        if "/" in t:
            t = id_to_locator_string_mapping[t]

        return f"{m.group(1)} `{t}` {m.group(3)}"

    def update_table(
        self,
        datastore: Datastore,
        output: Table,
        inputs: Dict[str, Table],

    ) -> None:
        referenced_tables_pattern = re.compile(
            r"(.*)(ref\(\"[a-zA-Z][a-zA-Z0-1._/-]*\"\))(.*)", re.MULTILINE
        )
        referenced_tables = [
            m[1][5:-2] for m in referenced_tables_pattern.findall(self.sql_code)
        ]
        assert set(referenced_tables) == set(inputs.keys()), f"{referenced_tables}, {inputs.keys()}"

        code = self.sql_code
        if inputs:
            id_to_locator_string = {
                t.id: datastore.locator_string(t) for t in inputs.values()
            }
            code = re.sub(
                referenced_tables_pattern,
                lambda match: self._substitute_table_names(id_to_locator_string, match),
                code,
            )

        logging.info(f"SQL code: \n{code}")
        datastore.write_from_sql_query(output, code, self.update_method)


class ParallelPandasJob(Job):
    """
    Allows to execute transformation in parallel on data groups defined by a given key.
    """

    def __init__(
        self,
        group_fn: Callable[[pd.DataFrame], pd.DataFrame],
        group_by_key: List[str],
        order_by_key: List[str] = [],
        clustering_key: List[str] = [],
        update_method: UpdateMethod = UpdateMethod.REPLACE,
        num_workers: int = 1,
        runner: str = "DataflowRunner",
    ):
        self.group_fn = group_fn
        self.group_by_key = group_by_key
        self.order_by_key = order_by_key
        self.clustering_key = clustering_key
        self.num_workers = num_workers
        self.update_method = update_method
        self.runner = runner

    def _field(self, name, type, mode="NULLABLE", fields=[]):
        field = {
            "name": name,
            "type": type,
            "mode": mode,
        }
        if fields:
            field["fields"] = fields
        return field

    def _get_bq_field(self, field: pa.Field):
        t = field.type
        if pa.types.is_boolean(t):
            return self._field(field.name, "BOOLEAN")
        if pa.types.is_integer(t):
            return self._field(field.name, "INTEGER")
        if pa.types.is_floating(t):
            return self._field(field.name, "FLOAT")
        if pa.types.is_decimal(t):
            return self._field(field.name, "NUMERIC")
        if pa.types.is_string(t):
            return self._field(field.name, "STRING")
        if pa.types.is_date(t):
            return self._field(field.name, "DATE")
        if pa.types.is_temporal(t):
            return self._field(field.name, "TIMESTAMP")
        if pa.types.is_struct(t):
            fields = [self._get_bq_field(f) for f in field]
            return self._field(field.name, "RECORD", fields=fields)
        if pa.types.is_list(t):
            res = self._get_bq_field(self, t.value_type)
            res["mode"] = "REPEATED"
            return res

        raise NotImplementedError(f"No conversion available to BigQuery for field: {field}")

    def _get_bq_schema(self, schema: pa.Schema) -> List[bigquery.SchemaField]:
        bq_fields = [self._get_bq_field(field) for field in schema]
        bq_schema = {
            "fields": bq_fields,
        }
        return bq_schema

    def update_table(
        self,
        datastore: Datastore,
        output: Table,
        inputs: Dict[str, Table]
    ) -> None:
        def extract_key(row):
            return [row[gkey] for gkey in self.group_by_key]

        def to_dataframe(group_key, group_value):
            group_df = pd.DataFrame.from_records(group_value)
            if self.order_by_key:
                group_df = group_df.sort_values(self.order_by_key)
            return group_df

        def to_dict(df):
            return df.to_dict(orient="records")

        output_ref = datastore.locator_string(output)
        job_id = output_ref.replace(".", "--").replace("_", "-")

        options = PipelineOptions.from_dictionary({
            "job_name": job_id,
            "project": datastore.default_project_id,
            "runner": self.runner,
            "temp_location": datastore.file_temp_location,
            "region": datastore.default_compute_region,
            "machine_type": "n1-standard-2",
            "num_workers": self.num_workers,
            "save_main_session": True,
            "autoscaling_algorithm": "NONE",
            "requirements_file": "/app/requirements.txt",
        })

        with beam.Pipeline(options=options) as p:
            assert len(inputs) == 1, "Only a single table as an input is suported"

            input_table = next(iter(inputs.values()))
            input_ref = datastore.locator_string(input_table)
            input_ref = input_ref.replace(".", ":", 1)

            additional_bq_parameters = {}
            if self.clustering_key:
                additional_bq_parameters={
                    "clustering": {
                        "fields": self.clustering_key,
                    }
                }

            write_disposition = beam.io.BigQueryDisposition.WRITE_TRUNCATE
            if self.update_method == UpdateMethod.APPEND:
                write_disposition = beam.io.BigQueryDisposition.WRITE_APPEND

            schema = self._get_bq_schema(output.schema)
            (
                p
                | f"Read BQ table" >> beam.io.ReadFromBigQuery(table=input_ref)
                | f"Group data by key" >> beam.GroupBy(extract_key)
                | f"Map to pandas.DataFrame" >> beam.MapTuple(to_dataframe)
                | f"Apply user defined transform" >> beam.Map(self.group_fn)
                | f"Transform to Dict" >> beam.FlatMap(to_dict)
                | f"Write to BQ table" >> beam.io.WriteToBigQuery(
                    table=output_ref.replace(".", ":", 1),
                    schema=schema,
                    create_disposition=beam.io.BigQueryDisposition.CREATE_IF_NEEDED,
                    write_disposition=write_disposition,
                    additional_bq_parameters=additional_bq_parameters,
                )
            )


class PandasJob(Job):
    """
    PandasJob integrates well with tables with flat columns. While working with tables with
    nested columns other job types are available.

    Details:
    The equivalent to a pandas DataFrame in Arrow is a Table. Both consist of a set of named
    columns of equal length. While pandas only supports flat columns, the Table also provides
    nested columns, thus it can represent more data than a DataFrame, so a full conversion is not
    always possible. Additional information is available here:
    https://arrow.apache.org/docs/python/pandas.html
    """

    def __init__(
        self,
        entry_point: Callable[[Dict[str, pd.DataFrame]], pd.DataFrame],
        update_method: UpdateMethod,
    ):
        self.entry_point = entry_point
        self.update_method = update_method

    def update_table(
        self,
        datastore: Datastore,
        output: Table,
        inputs: Dict[str, Table]
    ) -> None:
        input_dfs = {}
        for table_id, table in inputs.items():
            input_dfs[table_id] = datastore.read_as_arrow(table).to_pandas()

        df = self.entry_point(input_dfs)

        data = pa.Table.from_pandas(df, preserve_index=False, schema=output.schema)
        datastore.write_from_arrow(output, data, self.update_method)

==> peach-cli-0.0.12.tar.gz: peach-cli-0.0.12/src/peach/jobs.py <==
import os
from functools import cached_property
import logging
import re
from typing import Callable, Dict, List
import uuid

import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions

from google.cloud import bigquery
import pandas as pd
import pyarrow as pa

from peach.constants import UpdateMethod
from peach.datastores import Datastore
from peach.table import Table


class Job:
    def is_cli(self):
        return False

    def update_table(
        self,
        datastore: Datastore,
        output: Table,
        inputs: Dict[str, Table],
    ) -> None:
        raise NotImplementedError()


class SqlJob(Job):

    def __init__(self, sql_file: str, update_method: UpdateMethod):
        self.sql_file = sql_file
        self.update_method = update_method

    @cached_property
    def sql_code(self) -> str:
        with open(self.sql_file, "r") as fo:
            return fo.read()

    def _substitute_table_names(
        self, id_to_locator_string_mapping: Dict[str, str], m: re.Match
    ) -> str:
        t = m.group(2)[5:-2]
        assert "." in t or "/" in t, f"Table does not specify schema: {t}"

        if "/" in t:
            t = id_to_locator_string_mapping[t]

        return f"{m.group(1)} `{t}` {m.group(3)}"

    def update_table(
        self,
        datastore: Datastore,
        output: Table,
        inputs: Dict[str, Table],

    ) -> None:
        referenced_tables_pattern = re.compile(
            r"(.*)(ref\(\"[a-zA-Z][a-zA-Z0-1._/-]*\"\))(.*)", re.MULTILINE
        )
        referenced_tables = [
            m[1][5:-2] for m in referenced_tables_pattern.findall(self.sql_code)
        ]
        assert set(referenced_tables) == set(inputs.keys()), f"{referenced_tables}, {inputs.keys()}"

        code = self.sql_code
        if inputs:
            id_to_locator_string = {
                t.id: datastore.locator_string(t) for t in inputs.values()
            }
            code = re.sub(
                referenced_tables_pattern,
                lambda match: self._substitute_table_names(id_to_locator_string, match),
                code,
            )

        logging.info(f"SQL code: \n{code}")
        datastore.write_from_sql_query(output, code, self.update_method)


class ParallelPandasJob(Job):
    """
    Allows to execute transformation in parallel on data groups defined by a given key.
    """

    def __init__(
        self,
        group_fn: Callable[[pd.DataFrame], pd.DataFrame],
        group_by_key: List[str],
        order_by_key: List[str] = [],
        clustering_key: List[str] = [],
        update_method: UpdateMethod = UpdateMethod.REPLACE,
        num_workers: int = 1,
        runner: str = "DataflowRunner",
    ):
        self.group_fn = group_fn
        self.group_by_key = group_by_key
        self.order_by_key = order_by_key
        self.clustering_key = clustering_key
        self.num_workers = num_workers
        self.update_method = update_method
        self.runner = runner

    def _field(self, name, type, mode="NULLABLE", fields=[]):
        field = {
            "name": name,
            "type": type,
            "mode": mode,
        }
        if fields:
            field["fields"] = fields
        return field

    def _get_bq_field(self, field: pa.Field):
        t = field.type
        if pa.types.is_boolean(t):
            return self._field(field.name, "BOOLEAN")
        if pa.types.is_integer(t):
            return self._field(field.name, "INTEGER")
        if pa.types.is_floating(t):
            return self._field(field.name, "FLOAT")
        if pa.types.is_decimal(t):
            return self._field(field.name, "NUMERIC")
        if pa.types.is_string(t):
            return self._field(field.name, "STRING")
        if pa.types.is_date(t):
            return self._field(field.name, "DATE")
        if pa.types.is_temporal(t):
            return self._field(field.name, "TIMESTAMP")
        if pa.types.is_struct(t):
            fields = [self._get_bq_field(f) for f in field]
            return self._field(field.name, "RECORD", fields=fields)
        if pa.types.is_list(t):
            res = self._get_bq_field(self, t.value_type)
            res["mode"] = "REPEATED"
            return res

        raise NotImplementedError(f"No conversion available to BigQuery for field: {field}")

    def _get_bq_schema(self, schema: pa.Schema) -> List[bigquery.SchemaField]:
        bq_fields = [self._get_bq_field(field) for field in schema]
        bq_schema = {
            "fields": bq_fields,
        }
        return bq_schema

    def update_table(
        self,
        datastore: Datastore,
        output: Table,
        inputs: Dict[str, Table]
    ) -> None:
        def extract_key(row):
            return [row[gkey] for gkey in self.group_by_key]

        def to_dataframe(group_key, group_value):
            group_df = pd.DataFrame.from_records(list(group_value))
            if self.order_by_key:
                group_df = group_df.sort_values(self.order_by_key)
            return group_df

        def to_dict(df):
            return df.to_dict(orient="records")

        output_ref = datastore.locator_string(output)
        job_id = output_ref.replace(".", "--").replace("_", "-")

        opts = {
            "job_name": f"peach-{str(uuid.uuid4())}",
            "project": datastore.default_project_id,
            "runner": self.runner,
            "temp_location": datastore.file_temp_location,
            "region": datastore.default_compute_region,
            "machine_type": "n1-standard-2",
            "num_workers": self.num_workers,
            "save_main_session": True,
            "autoscaling_algorithm": "NONE",
        }
        if os.path.exists("/app/setup.py"):
            opts["setup_file"] = "/app/setup.py"
        else:
            opts["requirements_file"] = "/app/requirements.txt"
        options = PipelineOptions.from_dictionary(opts)

        with beam.Pipeline(options=options) as p:
            assert len(inputs) == 1, "Only a single table as an input is suported"

            input_table = next(iter(inputs.values()))
            input_ref = datastore.locator_string(input_table)
            input_ref = input_ref.replace(".", ":", 1)

            additional_bq_parameters = {}
            if self.clustering_key:
                additional_bq_parameters={
                    "clustering": {
                        "fields": self.clustering_key,
                    }
                }

            write_disposition = beam.io.BigQueryDisposition.WRITE_TRUNCATE
            if self.update_method == UpdateMethod.APPEND:
                write_disposition = beam.io.BigQueryDisposition.WRITE_APPEND

            schema = self._get_bq_schema(output.schema)
            (
                p
                | f"Read BQ table" >> beam.io.ReadFromBigQuery(table=input_ref)
                | f"Group data by key" >> beam.GroupBy(extract_key)
                | f"Map to pandas.DataFrame" >> beam.MapTuple(to_dataframe)
                | f"Apply user defined transform" >> beam.Map(self.group_fn)
                | f"Transform to Dict" >> beam.FlatMap(to_dict)
                | f"Write to BQ table" >> beam.io.WriteToBigQuery(
                    table=output_ref.replace(".", ":", 1),
                    schema=schema,
                    create_disposition=beam.io.BigQueryDisposition.CREATE_IF_NEEDED,
                    write_disposition=write_disposition,
                    additional_bq_parameters=additional_bq_parameters,
                )
            )


class PandasJob(Job):
    """
    PandasJob integrates well with tables with flat columns. While working with tables with
    nested columns other job types are available.

    Details:
    The equivalent to a pandas DataFrame in Arrow is a Table. Both consist of a set of named
    columns of equal length. While pandas only supports flat columns, the Table also provides
    nested columns, thus it can represent more data than a DataFrame, so a full conversion is not
    always possible. Additional information is available here:
    https://arrow.apache.org/docs/python/pandas.html
    """

    def __init__(
        self,
        entry_point: Callable[[Dict[str, pd.DataFrame]], pd.DataFrame],
        update_method: UpdateMethod,
    ):
        self.entry_point = entry_point
        self.update_method = update_method

    def update_table(
        self,
        datastore: Datastore,
        output: Table,
        inputs: Dict[str, Table]
    ) -> None:
        input_dfs = {}
        for table_id, table in inputs.items():
            input_dfs[table_id] = datastore.read_as_arrow(table).to_pandas()

        df = self.entry_point(input_dfs)

        data = pa.Table.from_pandas(df, preserve_index=False, schema=output.schema)
        datastore.write_from_arrow(output, data, self.update_method)


class CLIJob(Job):

    def __init__(self, args):
        self.args = args

    def is_cli(self):
        return True

    def update_table(
        self,
        datastore: Datastore,
        output: Table,
        inputs: Dict[str, Table]
    ) -> None:
        raise NotImplementedError("CLIJob runs the script directly")
